/* flow.c
 * flow core functions
 *
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@iserm.com>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>

#include "flow.h"
#include "proto.h"
#include "rule.h"
#include "log.h"
#include "dmemory.h"
#include "grp_flows.h"
#include "fthread.h"


/** define */
#define FW_TBL_ELEMENT_DELTA   10000

/** external variables */
extern prot_desc *prot_tbl;
extern int prot_tbl_dim;


/** internal variables */
static flow *volatile flow_tbl;   /* table of flows... all flows */
static volatile unsigned long tbl_dim;     /* dimension of table */
static volatile unsigned long flow_num;    /* number of open flows */
static pthread_mutex_t flow_mux;           /* mutex to access atomicly the tbl */
static volatile pthread_t ptrd_lock;       /* ptread that lock access */
static short nesting;                      /* lock nesting */
static unsigned long gbl_time;             /* flow global time */


/** internal functions */
static int FlowElemInit(flow *flw, bool reset)
{
    flw->elab = FALSE;
    flw->proto_id = -1;
    flw->fthd_id = -1;
    flw->pfid = -1;
    flw->grp_id = -1;
    flw->grp_fuse = FALSE;
    flw->stack = NULL;
    flw->name[0] = '\0';
    flw->time = ULONG_MAX;
    flw->ustime = ULONG_MAX;
    flw->pkt_q = NULL;
    flw->fpkt = NULL;
    flw->cpkt = NULL;
    flw->close = FALSE;
    flw->get_to = -1;
    if (reset == FALSE) {
        /* allocate mutex and inizialize it */
        flw->mux = xmalloc(sizeof(pthread_mutex_t));
        pthread_mutex_init(flw->mux, NULL);
        flw->cond = xmalloc(sizeof(pthread_cond_t));
        pthread_cond_init(flw->cond, NULL);
    }
    flw->pkt_num = 0;
    flw->priv_data_pn = NULL;
    flw->priv_data = NULL;
    flw->priv_data_prot = -1;
#ifdef XPL_PEDANTIC_STATISTICS
    flw->pkt_tot = 0;
#endif

    return 0;
}


static int FlowTblExtend(void)
{
    unsigned long i, len;
    flow *new;

    len = tbl_dim + FW_TBL_ELEMENT_DELTA;

    /* lock all mutex */
    for (i=0; i<tbl_dim; i++) {
        pthread_mutex_lock(flow_tbl[i].mux);
    }

    /* extend memory(copy) */
    new = xrealloc(flow_tbl, sizeof(flow)*(len));
    if (new == NULL)
        return -1;
    
    /* initialize new elements */
    for (i=tbl_dim; i<len; i++) {
        memset(&new[i], 0, sizeof(flow));
        FlowElemInit(&(new[i]), FALSE);
    }

    flow_tbl = new;
    /* unlock all mutex */
    for (i=0; i<tbl_dim; i++) {
        pthread_mutex_unlock(flow_tbl[i].mux);
    }
    tbl_dim = len;

    return 0;
}


static inline void FlowTblLock()
{
    if (pthread_mutex_trylock(&flow_mux) != 0) {
        if (ptrd_lock != pthread_self()) {
            pthread_mutex_lock(&flow_mux);
        }
    }
    ptrd_lock = pthread_self();
    nesting++;
}


static inline void FlowTblUnlock()
{
    nesting--;
    if (nesting == 0) {
        ptrd_lock = 0;
        pthread_mutex_unlock(&flow_mux);
    }
}


int FlowInit(void)
{ 
    flow_tbl = NULL;
    tbl_dim = 0;
    flow_num = 0;
    gbl_time = ULONG_MAX; /* no time */

    /* base flows tbl */
    if (FlowTblExtend() == -1) {
        LogPrintf(LV_ERROR, "Unable to inizialie flows data table");
        return -1;
    }
    pthread_mutex_init(&flow_mux, NULL);
    ptrd_lock = 0;
    nesting = 0;

    return 0;
}


const pstack_f* FlowStack(int flow_id)
{
    pstack_f* res;

    FlowTblLock();

    res = flow_tbl[flow_id].stack;

    FlowTblUnlock();

    return res;
}


int FlowCreate(pstack_f *stk)
{
    int i;

    FlowTblLock();

    /* search free location */
    for (i=0; i<tbl_dim; i++) {
        if (flow_tbl[i].stack == NULL) {
            break;
        }
    }
    if (i == tbl_dim) {
        if (FlowTblExtend() == -1) {
            LogPrintf(LV_ERROR, "Unable to extend flows data table");
            FlowTblUnlock();

            return -1;
        }
    }
#ifdef XPL_CHECK_CODE
    if (flow_tbl[i].fpkt != NULL || flow_tbl[i].pkt_num != 0) {
        LogPrintf(LV_OOPS, "flow elemnt free whit packet!");
    }
#endif
    FlowElemInit(&flow_tbl[i], TRUE);
    flow_tbl[i].stack = ProtCopyFrame(stk, TRUE);
    flow_tbl[i].stack->flow_id = i;
    flow_tbl[i].proto_id = ProtFrameProtocol(stk);
    flow_tbl[i].pfid = ProtParent(stk);
    flow_num++;

    FlowTblUnlock();

    return i;
}


int FlowClose(int flow_id)
{
    int ret;

    FlowTblLock();

    /* set close */
    if (flow_tbl[flow_id].close == TRUE) {
        FlowTblUnlock();
        
        return 0;
    }

    /* flush data */
    ProtFlushFlow(flow_tbl[flow_id].stack->pid, flow_id);

    pthread_mutex_lock(flow_tbl[flow_id].mux);

    flow_tbl[flow_id].close = TRUE;

    /* if this flow isn't in elaboration search an heuristic dissector */
    if (flow_tbl[flow_id].elab == FALSE) {
        pthread_mutex_unlock(flow_tbl[flow_id].mux);
        ret = ProtSearchHeuDisec(flow_tbl[flow_id].proto_id, flow_id);
        pthread_mutex_lock(flow_tbl[flow_id].mux);
    }

#ifndef XPL_CHECK_CODE
    if (flow_tbl[flow_id].pkt_num != 0 && flow_tbl[flow_id].fpkt == NULL) {
        LogPrintf(LV_OOPS, "bug in function %s line: %d", __FUNCTION__, __LINE__);
        while (1) {
            sleep(1);
        }
    }
#endif

    /* count the packet from protocol node to protocol dissector */
#ifdef XPL_PEDANTIC_STATISTICS
    if (flow_tbl[flow_id].proto_id != -1 &&  flow_tbl[flow_id].proto_id != flow_tbl[flow_id].pfid)
        ProtPktFromNode(flow_tbl[flow_id].proto_id, flow_tbl[flow_id].pkt_tot);
#endif

    /* wakeup flow in wait */
    pthread_cond_signal(flow_tbl[flow_id].cond);
    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    /* if this flow isn't in elaboration delete it */
    if (flow_tbl[flow_id].elab == FALSE) {
        if (flow_tbl[flow_id].pkt_num != 0)
            LogPrintf(LV_DEBUG, "FlowClose: flow %i no elab... delete it", flow_id);
        FlowDelete(flow_id);
    }

    FlowTblUnlock();

    return 0;
}


void FlowCloseAll(void)
{
    int i, cnt;
    int base;

    FlowTblLock();

    cnt = 0;
    base = flow_num;

    for (i=0; i<tbl_dim && cnt<base; i++) {
        if (flow_tbl[i].stack != NULL) {
            cnt++;
            FlowTblUnlock(); /* not necessary to FlowClose bat useful for scheduler thread */
            FlowClose(i);
            FlowTblLock();
        }
    }

    FlowTblUnlock();
}



int FlowDelete(int flow_id)
{
    int ret;
    packet *pkt, *tmp;
    int i, cnt, base;

    FlowTblLock();

    /* check if flow is closed */
    if (flow_tbl[flow_id].close == FALSE) {
        LogPrintf(LV_ERROR, "Delete a open flow");
        FlowClose(flow_id);
    }

    /* close all flow son of this */
    cnt = 0;
    base = flow_num;
    for (i=0; i<tbl_dim && cnt<base; i++) {
        if (flow_tbl[i].stack != NULL) {
            cnt++;
            if (flow_tbl[i].pfid == flow_id) {
                FlowClose(i);
                flow_tbl[i].pfid = -1; /* parent terminated */
            }
        }
    }

    /* check if flow have packet in the queue */
    pthread_mutex_lock(flow_tbl[flow_id].mux);
    if (flow_tbl[flow_id].pkt_num != 0) {
        if (flow_tbl[flow_id].elab == TRUE) {
            LogPrintf(LV_ERROR, "Deleted a flow with packet %d", flow_tbl[flow_id].pkt_num);
            ProtStackFrmDisp(FlowStack(flow_id), TRUE);
        }
        pkt = flow_tbl[flow_id].fpkt;
        
        /* free all packet */
#ifndef XPL_CHECK_CODE
        PktFree(pkt); /* recursive */
        pkt = NULL;
#else
        while (pkt != NULL) {
            tmp = pkt;
            pkt = pkt->next;
            tmp->next = NULL;
            PktFree(tmp);
            flow_tbl[flow_id].pkt_num--;
        }
        if (flow_tbl[flow_id].pkt_num != 0) {
            LogPrintf(LV_OOPS, "bug in pkt flow counter");
        }
#endif
    }
    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    /* remove from grp */
    if (flow_tbl[flow_id].grp_id != -1) {
        ret = GrpFlowRm(flow_tbl[flow_id].grp_id, flow_id);
#ifdef XPL_CHECK_CODE
        if (ret == -1) {
            LogPrintf(LV_OOPS, "bug in GrpFlow Add/Rm gestion and use");
        }
#endif
        if (GrpFlowEmpty(flow_tbl[flow_id].grp_id)) {
            GrpFlowDelete(flow_tbl[flow_id].grp_id);
        }
    }
    
    /* thread */
    if (flow_tbl[flow_id].elab == TRUE) {
        if (flow_tbl[flow_id].grp_id == -1) {
            ProtRunFlowDec(flow_tbl[flow_id].proto_id);
            FthreadChFlow(flow_tbl[flow_id].fthd_id, -1);
        }
        else {
            LogPrintf(LV_FATAL, "To be implemented (%s)", __FUNCTION__);
            exit(-1);
        }
    }
    
    /* stack */
    ProtDelFrame(flow_tbl[flow_id].stack);

    /* reset flow cel */
    FlowElemInit(flow_tbl+flow_id, TRUE);

    flow_num--;

    FlowTblUnlock();

    return 0;
}


packet* FlowGetPkt(int flow_id)
{
    packet *pkt;
    struct timespec to;

    pthread_mutex_lock(flow_tbl[flow_id].mux);
    
    while (flow_tbl[flow_id].fpkt == NULL) {
        if (flow_tbl[flow_id].close == TRUE)
            break;
        if (flow_tbl[flow_id].get_to > 0) {
            to.tv_sec = 0;
            to.tv_nsec = flow_tbl[flow_id].get_to*1000000;
            pthread_cond_timedwait(flow_tbl[flow_id].cond, flow_tbl[flow_id].mux, &to);
        }
        else if (flow_tbl[flow_id].get_to == -1)
            pthread_cond_wait(flow_tbl[flow_id].cond, flow_tbl[flow_id].mux);
    }
    if (flow_tbl[flow_id].fpkt == NULL) {
        pthread_mutex_unlock(flow_tbl[flow_id].mux);

        return NULL;
    }

    pkt = flow_tbl[flow_id].fpkt;
    flow_tbl[flow_id].fpkt = flow_tbl[flow_id].fpkt->next;
#ifdef XPL_CHECK_CODE
    if (pkt != flow_tbl[flow_id].cpkt) {
        LogPrintf(LV_OOPS, "bug in funcrion %s line: %d", __FUNCTION__, __LINE__);
        exit(-1);
    }
#endif
    flow_tbl[flow_id].cpkt = flow_tbl[flow_id].fpkt;
    if (flow_tbl[flow_id].fpkt == NULL) {
        flow_tbl[flow_id].pkt_q = NULL;
    }

    flow_tbl[flow_id].time = pkt->cap_sec;
    flow_tbl[flow_id].ustime = pkt->cap_usec;
    flow_tbl[flow_id].pkt_num--;
#ifdef XPL_CHECK_CODE
    if (flow_tbl[flow_id].pkt_num != 0 && flow_tbl[flow_id].fpkt == NULL) {
        LogPrintf(LV_OOPS, "bug in function %s line: %d", __FUNCTION__, __LINE__);
        exit(-1);
    }
#endif
    pthread_mutex_unlock(flow_tbl[flow_id].mux);
    pkt->next = NULL;

    return pkt;
}


packet* FlowGetPktCp(int flow_id)
{
    packet *pkt = NULL;

    pthread_mutex_lock(flow_tbl[flow_id].mux);

    if (flow_tbl[flow_id].cpkt != NULL) {
        pkt = PktCp(flow_tbl[flow_id].cpkt);
        flow_tbl[flow_id].cpkt = flow_tbl[flow_id].cpkt->next;
    }

    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    return pkt;
}


int FlowPktCpReset(int flow_id)
{
    pthread_mutex_lock(flow_tbl[flow_id].mux);

    flow_tbl[flow_id].cpkt = flow_tbl[flow_id].fpkt;

    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    return 0;
}


int FlowGetTimeOut(int flow_id, long ms)
{
    FlowTblLock();

    flow_tbl[flow_id].get_to = ms;

    FlowTblUnlock();

    return 0;
}


int FlowPutPkt(int flow_id, packet *pkt)
{
    pthread_mutex_lock(flow_tbl[flow_id].mux);

    if (flow_tbl[flow_id].pkt_q != NULL) {
        flow_tbl[flow_id].pkt_q->next = pkt;
        flow_tbl[flow_id].pkt_q = pkt;
    }
    else {
        flow_tbl[flow_id].pkt_q = pkt;
        flow_tbl[flow_id].fpkt = pkt;
        flow_tbl[flow_id].cpkt = pkt;
        flow_tbl[flow_id].time = pkt->cap_sec;
        flow_tbl[flow_id].ustime = pkt->cap_usec;
    }
    flow_tbl[flow_id].ins_time = pkt->cap_sec;
    flow_tbl[flow_id].ins_ustime = pkt->cap_usec;
    flow_tbl[flow_id].pkt_num++;
#ifdef XPL_PEDANTIC_STATISTICS
    flow_tbl[flow_id].pkt_tot++;
#endif

#ifdef XPL_CHECK_CODE
    if (pkt->len > pkt->raw_len && pkt->raw != NULL) {
        LogPrintf(LV_OOPS, "Data dimension error raw:%lu data:%lu", pkt->raw_len, pkt->len);
        ProtStackFrmDisp(pkt->stk, TRUE);
        exit(-1);
    }
#endif
    /* wakeup flow in wait */
    pthread_cond_signal(flow_tbl[flow_id].cond);
    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    return 0;
}


unsigned long FlowPktNum(int flow_id)
{
    unsigned long ret;

    pthread_mutex_lock(flow_tbl[flow_id].mux);

    ret = flow_tbl[flow_id].pkt_num;

    pthread_mutex_unlock(flow_tbl[flow_id].mux);

    return ret;
}


bool FlowIsClose(int flow_id)
{
    bool ret;

    FlowTblLock();

    ret = flow_tbl[flow_id].close;

    FlowTblUnlock();

    return ret;
}


bool FlowGrpIsEmpty(int flow_id)
{
    bool ret;

    FlowTblLock();
    
#warning "gestire i flussi a gruppi fusi"
    if (flow_tbl[flow_id].close == TRUE && flow_tbl[flow_id].fpkt == NULL)
        ret = TRUE;
    else
        ret = FALSE;
    
    FlowTblUnlock();

    return ret;
}


int FlowGrpId(int flow_id)
{
    int id;

    FlowTblLock();

    id = flow_tbl[flow_id].grp_id;

    FlowTblUnlock();
    
    return id;
}


int FlowGrpSet(int flow_id, int gflow_id)
{
    FlowTblLock();

    flow_tbl[flow_id].grp_id = gflow_id;

    FlowTblUnlock();

    return 0;
}


bool FlowIsEmpty(int flow_id)
{
    bool ret;

    FlowTblLock();
    
    if (flow_tbl[flow_id].close == TRUE && flow_tbl[flow_id].fpkt == NULL)
        ret = TRUE;
    else
        ret = FALSE;
    
    FlowTblUnlock();

    return ret;
}


int FlowSetElab(int flow_id, int fthd_id)
{
    FlowTblLock();

#ifdef XPL_CHECK_CODE
    if (flow_tbl[flow_id].elab == TRUE) {
        LogPrintf(LV_OOPS, "flow (%s) alrady in elaboration.", flow_tbl[flow_id].name);
        ProtStackFrmDisp(flow_tbl[flow_id].stack, TRUE);
    }
#endif

    flow_tbl[flow_id].elab = TRUE;
    flow_tbl[flow_id].fthd_id = fthd_id;

    FlowTblUnlock();

    return 0;
}


int FlowThreadId(int flow_id)
{
    int id;

    FlowTblLock();

    id = flow_tbl[flow_id].fthd_id;

    FlowTblUnlock();
    
    return id;
}


inline bool FlowCallSubDis(int flow_id, bool state)
{
    bool ret;
    
    ret = flow_tbl[flow_id].dis;
    flow_tbl[flow_id].dis = state;
    
    return ret;
}


int FlowSetName(int flow_id, int prot_id)
{
    FlowTblLock();

#ifdef XPL_CHECK_CODE
    if (flow_tbl[flow_id].name[0] != '\0') {
        LogPrintf(LV_OOPS, "flow (%d) have alrady a name (%s).", flow_id, flow_tbl[flow_id].name);
        ProtStackFrmDisp(flow_tbl[flow_id].stack, TRUE);
    }
#endif

    sprintf(flow_tbl[flow_id].name, "%s_%d", prot_tbl[prot_id].name, flow_id);
    flow_tbl[flow_id].proto_id = prot_id;

    FlowTblUnlock();

    return 0;
}

const char* FlowName(int flow_id)
{
    return flow_tbl[flow_id].name;
}



bool FlowInElabor(int flow_id)
{
    bool ret;

    FlowTblLock();

    ret = flow_tbl[flow_id].elab;

    FlowTblUnlock();

    return ret;
}


int FlowSearch(pstack_f *stk)
{
    return ProtStackSearchNode(stk);
}


void* FlowPrivGet(int flow_id)
{    
    return flow_tbl[flow_id].priv_data;
}


int FlowPrivPut(int flow_id, void *data)
{
    flow_tbl[flow_id].priv_data = data;
    return 0;
}


void *FlowNodePrivGet(int flow_id)
{
    return flow_tbl[flow_id].priv_data_pn;
}


int FlowNodePrivPut(int flow_id, void *data)
{
    flow_tbl[flow_id].priv_data_pn = data;
    return 0;
}


int FlowSetGblTime(unsigned long ftm)
{
    gbl_time = ftm;
    
    return 0;
}


unsigned long FlowGetGblTime(void)
{
    return gbl_time;
}


unsigned long FlowTime(int flow_id)
{
    unsigned long ret;

    FlowTblLock();

    ret = flow_tbl[flow_id].time;
    
    FlowTblUnlock();

    return ret;
}


unsigned long FlowTimeQ(int flow_id)
{
    unsigned long ret;

    FlowTblLock();

    ret = flow_tbl[flow_id].ins_time;
    
    FlowTblUnlock();

    return ret;
}


unsigned long FlowNumber(void)
{
    unsigned long ret;
#ifdef XPL_CHECK_CODE
    int i, cnt, base, closed;
#endif

    FlowTblLock();

    ret = flow_num;
#ifdef XPL_CHECK_CODE
    /* it atomic and so also ProtOpenFlow in this case */
    if (ret != ProtOpenFlow()) {
        cnt = 0;
        base = flow_num;
        closed = 0;
        for (i=0; i<tbl_dim && cnt<base; i++) {
            if (flow_tbl[i].stack != NULL) {
                cnt++;
                if (flow_tbl[i].close)
                    closed++;
            }
        }
        if (closed != flow_num) 
            LogPrintf(LV_OOPS, "Number of flows (%d) isn't the same of the sum of flows (%d)!", flow_num, closed);
    }
#endif

    FlowTblUnlock();

    return ret;
}


unsigned long FlowTblDim(void)
{
    return tbl_dim;
}
