/* flow.h
 *
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@iserm.com>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef __FLOW_H__
#define __FLOW_H__

#include <pthread.h>

#include "ftypes.h"
#include "istypes.h"
#include "packet.h"


/* define */
#define FLOW_NAME_SIZE  20
#define FLOW_RULE_PKT   "pkt."
#define FLOW_GROUP_DIM  20
#define FLOW_TO         240

/** flow  */
typedef struct _flow_t flow;
struct _flow_t {
    int proto_id;          /**< protocol id of flow: master protocol */
    volatile bool elab;    /**< elaboration started */
    volatile int fthd_id;  /**< flow thread id */
    int pfid;              /**< parent flow id (-1 no parent) */
    int grp_id;            /**< group id (-1 -> none) */
    bool grp_fuse;         /**< all flows in the group to be compare as the same flow time line */
    pstack_f *volatile stack;   /**< flow stack */
    char  name[FLOW_NAME_SIZE]; /**< flow name */
    unsigned long time;         /**< flow time sec in elaboration */
    unsigned long ustime;       /**< flow time usec in elaboration */
    unsigned long ins_time;     /**< flow time in sec -last packet- */
    unsigned long ins_ustime;   /**< flow time usec -last packet- */
    packet *volatile pkt_q;     /**< packet queue (fifo) */
    packet *volatile fpkt;      /**< first packet in queue */
    packet *volatile cpkt;      /**< cpy packet in queue for FlowGetPktCp */
    volatile bool close;        /**< input queue closed */
    long get_to;           /**< get timeout in milli-second (-1 -> no timeout) */
    pthread_mutex_t *mux;  /**< mutex to access packet queue */
    pthread_cond_t *cond;  /**< condiction for access packet queue */
    volatile unsigned long pkt_num; /**< number of packet in the flow queue */
    void *priv_data_pn;    /**< private data of protocol node that generate a flow */
    void *priv_data;       /**< private data of dissector flow */
    int priv_data_prot;    /**< protocol of private data */
    bool dis;              /**< dissector or subdissector running for this flow */
#ifdef XPL_PEDANTIC_STATISTICS
    unsigned long pkt_tot; /**< total number of packet */
#endif
};


/** flow descriptor*/
typedef struct _flow_d flow_d;
struct _flow_d {
    int fid;         /**< flow id in global table flaws (-1 if free) */
    int pfid;        /**< parent flow id (-1 no parent) */
    const pstack_f *stack; /**< flow stack (point in to stack of flow structure) */
};


/** flow functions */
int FlowInit(void);
const pstack_f* FlowStack(int flow_id);
int FlowCreate(pstack_f *stk);
int FlowClose(int flow_id);
void FlowCloseAll(void);
int FlowDelete(int flow_id);
packet* FlowGetPkt(int flow_id);
packet* FlowGetPktCp(int flow_id);
int FlowPktCpReset(int flow_id);
int FlowGetTimeOut(int flow_id, long ms);
int FlowPutPkt(int flow_id, packet *pkt);
unsigned long FlowPktNum(int flow_id);
bool FlowIsClose(int flow_id);
bool FlowIsEmpty(int flow_id);
bool FlowGrpIsEmpty(int flow_id);
int FlowGrpId(int flow_id);
int FlowGrpSet(int flow_id, int gflow_id);
int FlowSetElab(int flow_id, int fthd_id);
int FlowSetName(int flow_id, int prot_id);
const char* FlowName(int flow_id);
bool FlowInElabor(int flow_id);
int FlowSearch(pstack_f *stk);
void *FlowPrivGet(int flow_id);
int FlowPrivPut(int flow_id, void *data);
void *FlowNodePrivGet(int flow_id);
int FlowNodePrivPut(int flow_id, void *data);
int FlowSetGblTime(unsigned long ftime);
unsigned long FlowGetGblTime(void);
unsigned long FlowTime(int flow_id);

/* private functions */
unsigned long FlowNumber(void);
int FlowThreadId(int flow_id);
inline bool FlowCallSubDis(int flow_id, bool state);
unsigned long FlowTimeQ(int flow_id);
unsigned long FlowTblDim(void);

#endif /* __FLOW_H__ */
