/* log.c
 *
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@iserm.com>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "log.h"
#include "proto.h"
#include "flow.h"
#include "dmemory.h"
#include "fthread.h"


#define LOG_CORE_ID       -1   /* see make files */
#define LOG_CAPTURE_ID    -2
#define LOG_DISPATCH_ID   -3


static unsigned short core_mask = LV_DEFAULT;
static unsigned short dispatch_mask = LV_DEFAULT;
static unsigned short capture_mask = LV_DEFAULT;


static int LogXml(char *fname, int fid)
{
    const pstack_f *stack;
    char *xmlog;
    int fd;

    stack = NULL;
    if (fid != -1) {
        stack = FlowStack(fid);
    }

    if (stack != NULL) {
        /* xml flow */
        xmlog = ProtStackFrmXML(stack);
        
        /* save_xml_file */
        fd = open(fname, O_CREAT|O_RDWR);
        if (fd != -1) {
            write(fd, xmlog, strlen(xmlog));
            close(fd);
        }
        xfree(xmlog);
    }
    else
        return -1;

    return 0;
}


int LogPrintfPrt(int prot_id, unsigned short level, const char *format, ...)
{
    va_list argptr;
    char log[LV_LINE_MAX_DIM];
    char *buff;
    char xml_file[256];
    int len;
    static int xml_id = 1;
    int fid;

    buff = log;
    len = 0;

    /* flowid */
    fid = FthreadFlowId(pthread_self());
        
    /* prot_id >= 0 is for dissector
       prot_id = -1 for core
       prot_id = -2 for capture modules
       prot_id = -3 for dispatcer moduls
    */

    /* log mask */
    if (prot_id > -1) {
        /* protocol dissector mask */
        level &= ProtLogMask(prot_id);

        /* name */
        len = sprintf(buff, "[%s]-", ProtLogName(prot_id));
    }
    else if (prot_id == LOG_CORE_ID) {
        /* core mask */
        level &= core_mask;

        /* name */
        len = sprintf(buff, "[CORE]-");
        
    }
    else if (prot_id == LOG_CAPTURE_ID) {
        /* capture mask */
        level &= capture_mask;

        /* name */
        len = sprintf(buff, "[CAPT]-");
    }
    else if (prot_id == LOG_DISPATCH_ID) {
        /* dispatch mask */
        level &= dispatch_mask;

        /* name */
        len = sprintf(buff, "[DISP]-");
    }
    buff += len;
    
    /* file id and name */
    /* level */
    switch (level) {
    case 0:
        return  0;
        break;

    case LV_OOPS:
        sprintf(xml_file,  "%s/oops_%i_%lu.xml", ProtTmpDir(), xml_id, time(NULL));
        if (LogXml(xml_file, fid) == 0) {
            len = sprintf(buff, "OOPS: (%i) ", xml_id);
            xml_id++;
        }
        else
            len = sprintf(buff, "OOPS: ");
        break;

    case LV_FATAL:
        sprintf(xml_file,  "%s/fatal_%i_%lu.xml", ProtTmpDir(), xml_id, time(NULL));
        if (LogXml(xml_file, fid) == 0) {
            len = sprintf(buff, "FATAL: (%i) ", xml_id);
            xml_id++;
        }
        else
            len = sprintf(buff, "FATAL: ");
        break;
        
    case LV_ERROR:
        sprintf(xml_file,  "%s/error_%i_%lu.xml", ProtTmpDir(), xml_id, time(NULL));
        if (LogXml(xml_file, fid) == 0) {
            len = sprintf(buff, "ERROR: (%i) ", xml_id);
            xml_id++;
        }
        else
            len = sprintf(buff, "ERROR: ");
        break;
        
    case LV_WARNING:
        sprintf(xml_file,  "%s/warn_%i_%lu.xml", ProtTmpDir(), xml_id, time(NULL));
        if (LogXml(xml_file, fid) == 0) {
            len = sprintf(buff, "WARNING: (%i) ", xml_id);
            xml_id++;
        }
        else
            len = sprintf(buff, "WARNING: ");
        break;
        
    case LV_INFO:
        len = sprintf(buff, "INFO: ");
        break;
        
    case LV_TRACE:
        len = sprintf(buff, "TRACE: ");
        break;
        
    case LV_DEBUG:
        len = sprintf(buff, "DEBUG: ");
        break;
        
    case LV_START:
        len = sprintf(buff, "START: ");
        break;
        
    default:
        len = sprintf(buff, "---: ");
        break;
    }
    buff += len;

    va_start(argptr, format);
    len = vsprintf(buff, format, argptr);
    va_end(argptr);
    buff += len;

    puts(log);

    return 0;
}


int LogSetMask(int component, unsigned short mask)
{
    switch (component) {
    case LOG_CORE_ID:
        core_mask = mask;
        break;

    case LOG_CAPTURE_ID:
        capture_mask = mask;
        break;

    case LOG_DISPATCH_ID:
        dispatch_mask = mask;
        break;
        
    default:
        printf("Log mask error\n");
        exit(-1);
        break;
    }

    return 0;
}
