/* prot.c
 * protocol core function dissector
 *
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@iserm.com>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "proto.h"
#include "dmemory.h"
#include "log.h"
#include "ftypes.h"
#include "rule.h"
#include "fthread.h"
#include "xplxml.h"


/* local structures */
struct dis_arg {
    int fid;
    int pid;
};


/** external variables */
extern prot_desc *prot_tbl;
extern int prot_tbl_dim;


/** internal variables */

static packet* ProtSearchDisec(packet *pkt, int *prot_id);


static inline void ProtLock(int prot_id)
{
    if (pthread_mutex_trylock(&prot_tbl[prot_id].mux) != 0) {
        if (prot_tbl[prot_id].ptrd_lock != pthread_self()) {
            pthread_mutex_lock(&prot_tbl[prot_id].mux);
        }
    }
    prot_tbl[prot_id].ptrd_lock = pthread_self();
    prot_tbl[prot_id].nesting++;
}


static inline void ProtUnlock(int prot_id)
{
    prot_tbl[prot_id].nesting--;
    if (prot_tbl[prot_id].nesting == 0) {
        pthread_mutex_unlock(&prot_tbl[prot_id].mux);
    }
}


static int ProtFlowTimeout(int id)
{
    int elm_chk, i, fnum;
    flow_d *ftbld;
    unsigned long gbl_time, ftime, pftime;

    gbl_time = FlowGetGblTime();
    ProtLock(id);
    ftbld = prot_tbl[id].ftbl;
    elm_chk = 0;
    fnum = prot_tbl[id].flow_num;
    for (i=0; i<prot_tbl[id].ftbl_dim && elm_chk<fnum; i++) {
        if (ftbld[i].fid != -1 && FlowCallSubDis(ftbld[i].fid, FALSE) == FALSE) {
            ftime = FlowTimeQ(ftbld[i].fid);
            if (ftbld[i].pfid != -1)
                pftime = FlowTimeQ(ftbld[i].pfid);
            else
                pftime = gbl_time;
            if (pftime > ftime + FLOW_TO) {
                FlowClose(ftbld[i].fid);
            }
            elm_chk++;
        }
    }
    ProtUnlock(id);

    return 0;
}


static int ProtAddFlow(int prot_id, int flow_id)
{
    flow_d *new;
    int i, dim;
    const pstack_f *nxt;

    if (prot_tbl[prot_id].ftbl_dim == prot_tbl[prot_id].flow_num) {
        /* extend table */
        dim = prot_tbl[prot_id].ftbl_dim + PROT_FLOW_DESC_ADD;
        new = xrealloc(prot_tbl[prot_id].ftbl, sizeof(flow_d)*dim);
        if (new == NULL) {
            LogPrintf(LV_ERROR, "Unable to allocate memory for protocol %s", prot_tbl[prot_id].name);
            return -1;
        }
        for (i=prot_tbl[prot_id].ftbl_dim; i<dim; i++) {
            new[i].fid = -1;
            new[i].stack = NULL;
            new[i].pfid = -1;
        }
        i = prot_tbl[prot_id].ftbl_dim;
        prot_tbl[prot_id].ftbl = new;
        prot_tbl[prot_id].ftbl_dim = dim;
    }
    else {
        /* search free descriptor */
        i = 0;
        while (prot_tbl[prot_id].ftbl[i].fid != -1) {
            i++;
#ifdef XPL_CHECK_CODE
            if (i == prot_tbl[prot_id].ftbl_dim) {
                LogPrintf(LV_OOPS, "bug in funcrion %s line: %d", __FUNCTION__, __LINE__);
                return -1;
            }
#endif
        }
    }

    nxt = FlowStack(flow_id);
    if (nxt == NULL)
        return -1;

    prot_tbl[prot_id].ftbl[i].fid = flow_id;
    prot_tbl[prot_id].ftbl[i].stack = nxt;
    prot_tbl[prot_id].ftbl[i].pfid = ProtParent(nxt);

    prot_tbl[prot_id].flow_num++;

    return i;
}


static void* ProtThreadPkt(void *arg)
{
    int pid;
    int fid;
    packet *pkt;

    pid = ((struct dis_arg *)arg)->pid;
    fid = ((struct dis_arg *)arg)->fid;
    DMemFree(arg);
    FthreadSync();

    pkt = FlowGetPkt(fid);
    while (pkt != NULL) {
        ProtDissec(pid, pkt);
        pkt = FlowGetPkt(fid);
    }

    FlowDelete(fid);

    FthreadEnd();

    return NULL;
}


static void* ProtThreadFlow(void *arg)
{
    int pid, flow_pid;
    int fid, orig;
    int thrid;
    packet *pkt;
    FlowDissector  FlowDis;

    pid = ((struct dis_arg *)arg)->pid;
    flow_pid = pid;
    orig = fid = ((struct dis_arg *)arg)->fid;
    DMemFree(arg);
    FthreadSync();

    thrid = FlowThreadId(fid);
#ifdef XPL_CHECK_CODE
    if (thrid == -1) {
        LogPrintf(LV_OOPS, "Thread of flow (%d->%s) don't have thread id", fid, FlowName(fid));
        exit(-1);
    }
#endif

    FlowDis = prot_tbl[pid].FlowDis;
    do {
        pkt = FlowDis(fid);
        if (pkt != NULL) {
            pid = flow_pid;
            pkt = ProtSearchDisec(pkt, &pid);
            if (pid != -1) {
                ProtDissec(pid, pkt);
                pkt = NULL;
            }
        }
        if (pkt != NULL) {
            if (prot_tbl[flow_pid].DflSubDis != NULL)
                pkt = prot_tbl[flow_pid].DflSubDis(pkt);
            if (pkt != NULL)
                PktFree(pkt);
        }
        fid = FthreadFlow(thrid);
    } while (fid != -1 && FlowInElabor(fid) == TRUE && (FlowGrpIsEmpty(fid) == FALSE || FlowPrivGet(fid) != NULL));

    if (fid != -1 && FlowInElabor(fid) == TRUE)
        FlowDelete(fid);

    FthreadEnd();

    return NULL;
}


static int ProtThread(int prot_id, int flow_id)
{
    struct dis_arg *arg;
    int ret;

    if (prot_tbl[prot_id].GrpFlow != NULL) {
#warning "Grp flow to be implement"
        LogPrintf(LV_OOPS, "functionality 'GroupFlow' not implemented");
        exit(-1);
        FlowPktCpReset(flow_id); // e anche gli altri
        prot_tbl[prot_id].GrpFlow(flow_id, NULL, 0);
        FlowPktCpReset(flow_id);
    }

    arg = DMemMalloc(sizeof(struct dis_arg));
    arg->pid = prot_id;
    arg->fid = flow_id;
    if (prot_tbl[prot_id].FlowDis != NULL) {
        ret = FthreadCreate(flow_id, ProtThreadFlow, arg);
    }
    else {
        ret = FthreadCreate(flow_id, ProtThreadPkt, arg);
    }

    return ret;
}


int ProtSearchHeuDisec(int id, int flow_id)
{
    int i, ret;

    ret = -1;

    ProtLock(id);

    /* seach heurist dissector */
    for (i=0; i<prot_tbl[id].stbl_dim; i++) {
        if (prot_tbl[id].stbl[i].heu_dep != NULL) {
            /* ProtCheck is allayse different from NULL, see ProtHeuDep */
            FlowPktCpReset(flow_id);
            if (prot_tbl[id].stbl[i].heu_dep->ProtCheck(flow_id) == TRUE) {
                FlowPktCpReset(flow_id);
                ProtRunFlowInc(prot_tbl[id].stbl[i].id);
                
                /* set name of protocol */
                FlowSetName(flow_id, prot_tbl[id].stbl[i].id);
                
                /* launch thread */
                if (ProtThread(prot_tbl[id].stbl[i].id, flow_id) != 0) {
                    LogPrintf(LV_ERROR, "Unable to start thread (%d)", FthreadRunning());
                }
                ret = 0;
                break;
            }
            FlowPktCpReset(flow_id);
        }
    }

    ProtUnlock(id);

    return ret;
}


static packet* ProtSearchDisec(packet *pkt, int *prot_id)
{
    volatile int id = *prot_id;
    int i, j, ret, elm_chk;
    flow_d *ftbld;
    int flow_id, flow_par, flow_par_c;
    pstack_f *flame_stk;
    int flowd_id;
    bool create;
    int flow_num;

#ifdef XPL_CHECK_CODE
    if (pkt == NULL) {
        LogPrintf(LV_FATAL, "Packet NULL in %s", __FUNCTION__);
        exit(-1);
    }
    if (id == -1) {
        LogPrintf(LV_FATAL, "Protocol ID error", __FUNCTION__);
        ProtStackFrmDisp(pkt->stk, TRUE);
        exit(-1);
    }
#endif

    *prot_id = -1;
    if (prot_tbl[id].flow == FALSE) {
        /* packet dissector */
        for (i=0; i<prot_tbl[id].stbl_dim; i++) {
            if (FTCmp(&(pkt->stk->attr[prot_tbl[id].stbl[i].sfpaid]),
                      &(prot_tbl[id].stbl[i].dep->val), prot_tbl[id].stbl[i].dep->type,
                      prot_tbl[id].stbl[i].dep->op) == 0) {
                id = prot_tbl[id].stbl[i].id;
                *prot_id = id;
                break;
            }
        }
    }
    else {
        /* this dissector is a node of flows and so it is a flow dissector */
        /* search flow if it exist that match prtocol rule */
        /* actual flow */
        pkt->stk->flow = TRUE;
        flame_stk = pkt->stk->pfp;
        flow_par = -1;
        while (flame_stk != NULL && flame_stk->flow == FALSE) {
            flame_stk = flame_stk->pfp;
        }
        if (flame_stk != NULL)
            flow_par = flame_stk->flow_id;
        flow_id = -1;
        flowd_id = -1;
        elm_chk = 0;
        ProtLock(id);
        ftbld = prot_tbl[id].ftbl;
        flow_num = prot_tbl[id].flow_num; /* SubDis can close a flow */
        create = TRUE;
        for (i=0; i<prot_tbl[id].ftbl_dim && elm_chk<flow_num; i++) {
            if (ftbld[i].fid != -1) {
                if (FlowIsClose(ftbld[i].fid) == FALSE) {
                    /* check if same parent flow */
                    flame_stk = ftbld[i].stack->pfp;
                    flow_par_c = -1;
                    while (flame_stk != NULL && flame_stk->flow == FALSE) {
                        flame_stk = flame_stk->pfp;
                    }
                    if (flame_stk != NULL)
                        flow_par_c = flame_stk->flow_id;
                    if (flow_par_c == flow_par) {
                        /* check rule */
                        ret = -1;
                        j = 0;
                        while (ret == -1 && j<prot_tbl[id].rule_num) {
                            ret = RuleCheck(&prot_tbl[id].rule[j++], ftbld[i].stack, pkt->stk);
                        }
                        if (ret == 1) {
                            flowd_id = i;
                            flow_id = ftbld[i].fid;
                            create = FALSE;
                            pkt->stk->flow_id = flow_id;
                            if (prot_tbl[id].SubDis != NULL) {
                                FlowCallSubDis(flow_id, TRUE);
                                prot_tbl[id].SubDis(flow_id, pkt);
                                /* FlowClose can be used in SubFlow and theme call ProtFlushFlow */
                                if (ftbld[flowd_id].fid == -1) {
                                    flow_id = -1;
                                    flowd_id = -1;
                                }
                            }
                            else {
                                FlowPutPkt(flow_id, pkt);
                            }
                            pkt = NULL;
                            break;
                        }
                    }
                }
                elm_chk++;
            }
        }
        if (create == TRUE) {
            /* create new flow */
            flow_id = FlowCreate(pkt->stk);
            if (flow_id != -1) {
                /* flow descriptor */
                pkt->stk->flow_id = flow_id;
                flowd_id = ProtAddFlow(id, flow_id);
                if (flowd_id != -1) {
                    if (prot_tbl[id].SubDis != NULL) {
                        prot_tbl[id].SubDis(flow_id, pkt);
                        flow_id = prot_tbl[id].ftbl[flowd_id].fid; /* if a a SubDis close flow */
                    }
                    else {
                        FlowPutPkt(flow_id, pkt);
                    }
                    pkt = NULL;
                }
                else {
                    LogPrintf(LV_ERROR, "Unable to add new flows in protocol %s", prot_tbl[id].name);
                    ProtStackFrmDisp(pkt->stk, TRUE);
                    FlowClose(flow_id);
                    FlowDelete(flow_id);
                    flow_id = -1;
                }
            }
            else {
                LogPrintf(LV_ERROR, "Unable to create new flaw");
            }
        }
        /* flow running */
        if (flow_id != -1) {
            ftbld = prot_tbl[id].ftbl;
#ifdef XPL_CHECK_CODE
            if (flowd_id == -1) {
                LogPrintf(LV_OOPS, "flow descriptor don't exist.");
                ProtStackFrmDisp(FlowStack(flow_id), TRUE);
                exit(-1);
            }
            if (flow_id != ftbld[flowd_id].fid) {
                LogPrintf(LV_OOPS, "flow descriptor 'differ' from flow ID. (%d: %d, %d->%d) ",
                          id, flow_id,  flowd_id, ftbld[flowd_id].fid);
                ProtStackFrmDisp(FlowStack(flow_id), TRUE);
                exit(-1);
            }
#endif
            if (FlowInElabor(flow_id) == FALSE) {
                create = FALSE;
                /* seach dissector */
                for (i=0; i<prot_tbl[id].stbl_dim; i++) {
                    if (prot_tbl[id].stbl[i].dep != NULL &&
                        FTCmp(&(ftbld[flowd_id].stack->attr[prot_tbl[id].stbl[i].sfpaid]), 
                              &(prot_tbl[id].stbl[i].dep->val), prot_tbl[id].stbl[i].dep->type,
                              prot_tbl[id].stbl[i].dep->op) == 0) {
                        /* if exist flow check function */
                        create = TRUE;
                        if (prot_tbl[id].stbl[i].dep->ProtCheck != NULL) {
                            FlowPktCpReset(flow_id);
                            create = prot_tbl[id].stbl[i].dep->ProtCheck(flow_id);
                            FlowPktCpReset(flow_id);
                        }
                        if (create == TRUE) {
                            ProtRunFlowInc(prot_tbl[id].stbl[i].id);

                            /* set name of protocol */
                            FlowSetName(flow_id, prot_tbl[id].stbl[i].id);

                            /* launch thread */
                            if (ProtThread(prot_tbl[id].stbl[i].id, flow_id) != 0) {
                                LogPrintf(LV_ERROR, "Unable to start thread (%d)", FthreadRunning());

                            }
                            break;
                        }
                    }
                }
                /* seach heurist dissector */
                if (create == FALSE) {
                    for (i=0; i<prot_tbl[id].stbl_dim; i++) {
                        if (prot_tbl[id].stbl[i].heu_dep != NULL) {
                            /* ProtCheck is allayse different from NULL, see ProtHeuDep */
                            FlowPktCpReset(flow_id);
                            if (prot_tbl[id].stbl[i].heu_dep->ProtCheck(flow_id) == TRUE) {
                                FlowPktCpReset(flow_id);
                                create = TRUE;
                                ProtRunFlowInc(prot_tbl[id].stbl[i].id);

                                /* set name of protocol */
                                FlowSetName(flow_id, prot_tbl[id].stbl[i].id);
                                
                                /* launch thread */
                                if (ProtThread(prot_tbl[id].stbl[i].id, flow_id) != 0) {
                                    LogPrintf(LV_ERROR, "Unable to start thread (%d)", FthreadRunning());
                                }
                                break;
                            }
                            FlowPktCpReset(flow_id);
                        }
                    }
                }
            }
        }
        ProtUnlock(id);
    }
    
    return pkt;
}


int ProtDissec(int prot_id, packet *pkt)
{
    PktDissector PktDis;
    packet *pkt_res;
    int id, pre_prot;

#ifdef XPL_CHECK_CODE
    if (pkt == NULL) {
        LogPrintf(LV_FATAL, "Packet NULL in %s", __FUNCTION__);
        exit(-1);
    }
#endif

    /* find dissector function */
    id = prot_id;
    PktDis = prot_tbl[id].PktDis;

    /* run */
    while (PktDis != NULL && pkt != NULL) {
#ifdef XPL_PEDANTIC_STATISTICS
        pthread_mutex_lock(&prot_tbl[id].cnt_mux);
        prot_tbl[id].pkt_tot++;
        pthread_mutex_unlock(&prot_tbl[id].cnt_mux);
#endif
        pkt_res = PktDis(pkt);
        if (pkt_res != NULL) {
            /* search new dissector */
            PktDis = NULL;
            pre_prot = id;
            pkt_res = ProtSearchDisec(pkt_res, &id);
            if (id != -1) {
                if (prot_tbl[id].tver == TRUE) {
                    prot_tbl[id].tver = FALSE;
                    ProtFlowTimeout(id);
                }
                PktDis = prot_tbl[id].PktDis;
            }
        }
        pkt = pkt_res;
    }
    if (prot_tbl[prot_id].tver == TRUE) {
        prot_tbl[prot_id].tver = FALSE;
        ProtFlowTimeout(prot_id);
    }
    /* if don't exist dissector or in error condiction */
    if (pkt != NULL) {
        if (prot_tbl[pre_prot].DflSubDis != NULL)
            pkt = prot_tbl[pre_prot].DflSubDis(pkt);
        if (pkt != NULL) {
            //ProtStackFrmDisp(pkt->stk, TRUE);
            PktFree(pkt);
        }
    }

    return 0;
}


int ProtFlushFlow(int prot_id, int flow_id)
{
    int i, j, ret;

    ProtLock(prot_id);

    /* check flow descriptor */
    for (i=0, j=0; i<prot_tbl[prot_id].ftbl_dim && j<prot_tbl[prot_id].flow_num ; i++) {
        if (prot_tbl[prot_id].ftbl[i].fid != -1) {
            if (prot_tbl[prot_id].ftbl[i].fid == flow_id)
                break;
            j++;
        }
    }
    
    if (j != prot_tbl[prot_id].flow_num) {
        ret = 0;
        if (prot_tbl[prot_id].SubDis != NULL)
            prot_tbl[prot_id].SubDis(flow_id, NULL);

        /* flush is execute at close of flow the remove flow descritpor */
        prot_tbl[prot_id].ftbl[i].fid = -1;
        prot_tbl[prot_id].ftbl[i].stack = NULL;
        prot_tbl[prot_id].ftbl[i].pfid = -1;
        prot_tbl[prot_id].flow_num--;
    }
    else {
        LogPrintf(LV_ERROR, "Flow %s is not in protocol %s", FlowName(flow_id), prot_tbl[prot_id].name);
        ret = -1;
    }    

    ProtUnlock(prot_id);

    return ret;
}


int ProtOpenFlow(void)
{
    int i, res;

    /* it is not atomic! */
    res = 0;
    for (i=0; i<prot_tbl_dim; i++) {
        res += prot_tbl[i].flow_num;
    }

    return res;
}


int ProtId(char *name)
{
    int i;

    /* search protocol */
    for (i=0; i<prot_tbl_dim; i++) {
        if (strcmp(name, prot_tbl[i].name) == 0)
            break;
    }
    if (i == prot_tbl_dim)
        return -1;

    return i;
}


int ProtAttrId(int pid, char *attr)
{
    int i;
    
    if (pid == -1)
        return -1;

    /* search attribute id */
    for (i=0; i<prot_tbl[pid].info_num; i++) {
        if (strcmp(attr, prot_tbl[pid].info[i].abbrev) == 0)
            break;
    }
    if (i == prot_tbl[pid].info_num)
        return -1;

    return i;
}


pstack_f* ProtCreateFrame(int prot_id)
{
    pstack_f *frame;
    int size;

    size = prot_tbl[prot_id].pstack_sz;
    frame = DMemMalloc(size);
    if (frame == NULL)
        return NULL;

    memset(frame, 0, size);
    frame->pid = prot_id;
    frame->flow = prot_tbl[prot_id].flow;
    frame->flow_id = -1;
    frame->pfp = NULL;
    
    return frame;
}


pstack_f* ProtCopyFrame(const pstack_f *stk, bool all)
{
    const pstack_f *next;
    pstack_f *frame, *new;
    int size, i, nv;

    new = NULL;
    frame = NULL;
    next = stk;
    do {
        size = prot_tbl[next->pid].pstack_sz;
        if (new == NULL) {
            frame = DMemMalloc(size);
            new = frame;
        }
        else {
            frame->pfp = DMemMalloc(size);
            frame = frame->pfp;
        }
        if (frame == NULL) {
            frame = new;
            while (frame != NULL) {
                new = frame->pfp;
                DMemFree(frame);
                frame = new;
            }
                
            return NULL;
        }
        
        memcpy(frame, next, sizeof(pstack_f));
        nv = prot_tbl[next->pid].info_num;
        for (i=0; i<nv; i++) {
            FTCopy(frame->attr+i, next->attr+i, prot_tbl[next->pid].info[i].type);
        }
        frame->pfp = NULL;

        next = next->pfp;
    } while (next != NULL && all == TRUE); 

    return new;
}


int ProtDelFrame(pstack_f *stk)
{
    pstack_f *nxt, *tmp;
    int nv, i;

    if (stk == NULL)
        return 0;
    
    nxt = stk;
    while (nxt != NULL) {
        nv = prot_tbl[nxt->pid].info_num;
        for (i=0; i<nv; i++) {
            FTFree(nxt->attr+i, prot_tbl[nxt->pid].info[i].type);
        }
        tmp = nxt;
        nxt = nxt->pfp;
        DMemFree(tmp);
    }

    return 0;
}


int ProtInsAttr(pstack_f *frame, int id, ftval *val)
{
    if (frame == NULL)
        return -1;
    
    FTCopy(&frame->attr[id], val,  prot_tbl[frame->pid].info[id].type);

    return 0;
}


int ProtGetAttr(const pstack_f *frame, int id, ftval *val)
{
    if (frame == NULL)
        return -1;
    
    FTCopy(val, &frame->attr[id], prot_tbl[frame->pid].info[id].type);

    return 0;
}


int ProtPeiComptId(int pid, char *abbrev)
{
    int i;
    
    /* search component id */
    for (i=0; i<prot_tbl[pid].peic_num; i++) {
        if (strcmp(abbrev, prot_tbl[pid].peic[i].abbrev) == 0)
            break;
    }
    if (i == prot_tbl[pid].peic_num)
        return -1;

    return i;
}


inline int ProtFrameProtocol(const pstack_f *frame)
{
    return frame->pid;
}


inline const pstack_f* ProtGetNxtFrame(const pstack_f *frame)
{
    return frame->pfp;
}


inline int ProtSetNxtFrame(pstack_f *frame, pstack_f *nxt)
{
    frame->pfp = nxt;
    
    return 0;
}


void ProtStackFrmDisp(const pstack_f *frame, bool all)
{
    const pstack_f *next;
    int i, num;
    char buff[512];
    int space = 0;
    char *resp[] = {"no", "yes"};
    
    num = 0;
    next = frame;
    while (next != NULL) {
        LogPrintf(LV_INFO, "\033[%liCframe %d - prot: %d,  flow: %s, id: %d -",
                  space, num, next->pid, resp[next->flow], next->flow_id);
        space  += 3;
        for (i=0; i<prot_tbl[next->pid].info_num; i++) {
            LogPrintf(LV_INFO, "\033[%liC%s: %s", space, prot_tbl[next->pid].info[i].abbrev,
                      FTString(&next->attr[i], prot_tbl[next->pid].info[i].type, buff));
        }
        if (all == TRUE)
            next = next->pfp;
        num++;
    }
}


char *ProtStackFrmXML(const pstack_f *frame)
{
    char *xml_buf;
    int dim, len, i, j;
    const pstack_f *next;
    char buff[512];

    i = 0;
    dim = 1024;
    xml_buf = xmalloc(dim);
    strcpy(xml_buf, XPL_HEADER);
    len = strlen(XPL_HEADER);
    len += sprintf(xml_buf+len, XPL_FLOW_OPEN, i);
    i++;
    next = frame;
    while (next != NULL) {
        if (len > dim-4096) {
            dim += 4096;
            xml_buf = xrealloc(xml_buf, dim);
        }
        len += sprintf(xml_buf+len, XPL_FRAME_OPEN,  prot_tbl[next->pid].name);
        for (j=0; j<prot_tbl[next->pid].info_num; j++) {
            len += sprintf(xml_buf+len, XPL_PROP, prot_tbl[next->pid].info[j].abbrev,
                           FTString(&next->attr[j], prot_tbl[next->pid].info[j].type, buff));
        }
        strcpy(xml_buf+len, XPL_FRAME_CLOSE);
        len += strlen(XPL_FRAME_CLOSE);
        next = next->pfp;
    }
    strcpy(xml_buf+len, XPL_FLOW_CLOSE);
    len += strlen(XPL_FLOW_CLOSE);

    return xml_buf;
}


PktDissector ProtPktDis(int prot_id)
{
    return prot_tbl[prot_id].PktDis;
}


FlowDissector ProtFlowDis(int prot_id)
{
    return prot_tbl[prot_id].FlowDis;
}


int ProtStackSearchNode(pstack_f *stk)
{
    int i, j, p_id;
    int elm_chk;
    flow_d *ftbld;
    pstack_f * flame_stk;
    int flow_id, flow_par, flow_par_c;
    int ret;

    if (stk == NULL && stk->flow == FALSE)
        return -1;

    /* node of flow parent of this stack */
    flame_stk = stk->pfp;
    flow_par = -1;
    while (flame_stk != NULL && flame_stk->flow == FALSE) {
        flame_stk = flame_stk->pfp;
    }
    if (flame_stk != NULL)
        flow_par = flame_stk->flow_id;

    p_id = stk->pid;
    elm_chk = 0;
    flow_id = -1;
    ProtLock(p_id);
    ftbld = prot_tbl[p_id].ftbl;
    for (i=0; i<prot_tbl[p_id].ftbl_dim && elm_chk<prot_tbl[p_id].flow_num; i++) {
        if (ftbld[i].fid != -1) {
            if (FlowIsClose(ftbld[i].fid) == FALSE && FlowInElabor(ftbld[i].fid) == FALSE) {
                /* check if same parent flow */
                flame_stk = ftbld[i].stack->pfp;
                flow_par_c = -1;
                while (flame_stk != NULL && flame_stk->flow == FALSE) {
                    flame_stk = flame_stk->pfp;
                }
                if (flame_stk != NULL)
                    flow_par_c = flame_stk->flow_id;
                if (flow_par_c == flow_par) {
                    /* check rule */
                    ret = -1;
                    j = 0;
                    while (ret == -1 && j<prot_tbl[p_id].rule_num) {
                        ret = RuleCheck(&prot_tbl[p_id].rule[j++], ftbld[i].stack, stk);
                    }
                    if (ret == 1) {
                        flow_id = ftbld[i].fid;
                        break;
                    }
                }
            }
            elm_chk++;
        }
    }
    ProtUnlock(p_id);

    return flow_id;
}


const pstack_f* ProtStackSearchProt(pstack_f *stk, int pid)
{
    while (stk != NULL && stk->pid != pid)
        stk = stk->pfp;

    return stk;
}


int ProtParent(const pstack_f *stk)
{
    const pstack_f *nxt;
    int i;

    i = -1;
    nxt = stk->pfp;
    while (nxt != NULL) {
        if (nxt->flow == TRUE) {
            i = nxt->flow_id;
            break;
        }
        nxt = nxt->pfp;
    }
    
    return i;
}


int ProtRunFlowInc(int prot_id)
{
    ProtLock(prot_id);

    prot_tbl[prot_id].flow_run++;
    prot_tbl[prot_id].flow_tot++;

    ProtUnlock(prot_id);

    return 0;
}


int ProtRunFlowDec(int prot_id)
{
    ProtLock(prot_id);

    prot_tbl[prot_id].flow_run--;
    
    ProtUnlock(prot_id);

    return 0;
}


int ProtRunningFlow(int prot_id)
{
    return prot_tbl[prot_id].flow_run;
}


unsigned long ProtTotFlow(int prot_id)
{
    return prot_tbl[prot_id].flow_tot;
}


int ProtStatus(void)
{
    int i;

    for (i=0; i<prot_tbl_dim; i++) {
#ifdef XPL_PEDANTIC_STATISTICS
        printf("%s: running: %i/%lu, subflow:%i/%i, tot pkt:%lu\n",
               prot_tbl[i].name, prot_tbl[i].flow_run, prot_tbl[i].flow_tot,
               prot_tbl[i].flow_num, prot_tbl[i].ftbl_dim, prot_tbl[i].pkt_tot);
#else
        printf("%s: running: %i/%lu, subflow:%i/%i\n",
               prot_tbl[i].name, prot_tbl[i].flow_run, prot_tbl[i].flow_tot,
               prot_tbl[i].flow_num, prot_tbl[i].ftbl_dim);
#endif
    }
    
    return 0;
}


unsigned short ProtLogMask(int prot_id)
{
    return prot_tbl[prot_id].log_mask;
}


const char* ProtLogName(int prot_id)
{
    return prot_tbl[prot_id].name;
}


#ifdef XPL_PEDANTIC_STATISTICS
int ProtPktFromNode(int prot_id, unsigned long pkt_tot)
{
    pthread_mutex_lock(&prot_tbl[prot_id].cnt_mux);
    prot_tbl[prot_id].pkt_tot += pkt_tot;
    pthread_mutex_unlock(&prot_tbl[prot_id].cnt_mux);
    
    return 0;
}

#endif
