/* dispatch.c
 * Dispatch interface
 *
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@iserm.com>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <pthread.h>

#include "dispatch.h"
#include "config_param.h"
#include "dmemory.h"
#include "log.h"
#include "pei.h"
#include "proto.h"


static void *handle;
static bool parallel;
static int (*DispInit)(void);
static int (*DispInsPei)(pei *ppei);


static void* DispatchAgent(void *arg)
{
    return NULL;
}


int DispatchInit(const char *file_cfg)
{
    FILE *fp;
    char module_dir[CFG_LINE_MAX_SIZE];
    char buffer[CFG_LINE_MAX_SIZE];
    char bufcpy[CFG_LINE_MAX_SIZE];
    char module_path[CFG_LINE_MAX_SIZE];
    char module_name[CFG_LINE_MAX_SIZE];
    char mask[CFG_LINE_MAX_SIZE];
    char *param;
    unsigned short logm;
    int res, nl, val;
    pthread_t pid;

    /* default */
    parallel = FALSE;

    /* find directory location of module from config file */
    fp = fopen(file_cfg, "r");
    if (fp == NULL) {
        LogPrintf(LV_ERROR, "Config file can't be opened");
        return -1;
    }
    module_dir[0] = '\0';
    module_name[0] = '\0';
    nl = 0;
    while (fgets(buffer, CFG_LINE_MAX_SIZE, fp) != NULL) {
        nl++;
        /* check all line */
        if (strlen(buffer)+1 == CFG_LINE_MAX_SIZE) {
            LogPrintf(LV_ERROR, "Config file line more length to %d characters", CFG_LINE_MAX_SIZE);
            return -1;
        }
        /* check if line is a comment */
        if (!CfgParIsComment(buffer)) {
            /* modules directory */
            param = strstr(buffer, CFG_PAR_MODULES_DIR);
            if (param != NULL) {
                if (module_dir[0] != '\0') {
                    LogPrintf(LV_ERROR, "Config param error: param '%s' defined two times", CFG_PAR_MODULES_DIR);
                    return -1;
                }
                res = sscanf(param, CFG_PAR_MODULES_DIR"=%s %s", module_dir, bufcpy);
                if (res > 0) {
                    if (res == 2 && !CfgParIsComment(bufcpy)) {
                        LogPrintf(LV_ERROR, "Config param error in line %d. Unknow param: %s", nl, bufcpy);
                        return -1;
                    }
                }
            }
            /* dispatcher module name */
            param = strstr(buffer, CFG_PAR_DISPATCH"=");
            if (param != NULL) {
                if (module_name[0] != '\0') {
                    LogPrintf(LV_ERROR, "Config param error: param '%s' defined two times", CFG_PAR_DISPATCH);
                    return -1;
                }
                res = sscanf(param, CFG_PAR_DISPATCH"=%s %s", module_name, bufcpy);
                if (res > 0) {
                    if (res == 2 && !CfgParIsComment(bufcpy)) {
                        /* log mask */
                        res = strncmp(bufcpy, CFG_PAR_MODULE_LOG, strlen(CFG_PAR_MODULE_LOG));
                        if (res != 0) {
                            LogPrintf(LV_ERROR, "Config param error in line %d. Unknow param: %s", nl, bufcpy);
                            return -1;
                        }
                        param = strstr(buffer, CFG_PAR_MODULE_LOG);
                        res = sscanf(param, CFG_PAR_MODULE_LOG"=%s %s", mask, bufcpy);
                        logm = LV_BASE;
                        if (res > 0) {
                            if (res == 2 && !CfgParIsComment(bufcpy)) {
                                LogPrintf(LV_ERROR, "Config param error in line %d. Unknow param: %s", nl, bufcpy);
                                return -1;
                            }
                            logm |= CfgParLogMask(mask, nl);
                        }
                        else {
                            LogPrintf(LV_ERROR, "Config param error in line %d. Unknow param: %s", nl, buffer);
                            return -1;
                        }
                        /* set mask */
                        LogSetMask(LOG_COMPONENT, logm);
                    }
                }
            }

            /* parallel o serial insert */
            param = strstr(buffer, CFG_PAR_DISPATCH_PARAL);
            if (param != NULL) {
                res = sscanf(param, CFG_PAR_DISPATCH_PARAL"=%i %s", &val, bufcpy);
                if (res > 0) {
                    if (res == 2 && !CfgParIsComment(bufcpy)) {
                        LogPrintf(LV_ERROR, "Config param error in line %d. Unknow param: %s", nl, bufcpy);
                        return -1;
                    }
                    else {
                        if (val == 1)
                            parallel = TRUE;
                        else
                            parallel = FALSE;
                    }
                }
            }
        }
    }
    fclose(fp);

    /* check name */
    if (module_name[0] == '\0') {
        LogPrintf(LV_WARNING, "The dispatch module isn't defined, will be used 'none' dispatch module");
        printf("The dispatch module isn't defined, will be used 'none' dispatch module\n");

        /* default dispatcher */
        strcpy(module_name, "disp_none.so");
    }

    /* module path */
    sprintf(module_path, "%s/%s", module_dir, module_name);

    /* open module */
    handle = dlopen(module_path, RTLD_NOW);
    if (handle == NULL) {
        printf("Can't load dispatch module %s\n", dlerror());
        return -1;
    }
    
    /* attach funztions */
    DispInit = dlsym(handle, DISP_INIT_FUN);
    if (DispInit == NULL) {
        printf("Dispatch module don't contain function %s\n", DISP_INIT_FUN);
        return -1;
    }

    DispInsPei = dlsym(handle, DISP_INDPEI_FUN);
    if (DispInsPei == NULL) {
        printf("Dispatch module don't contain function %s\n", DISP_INDPEI_FUN);
        return -1;
    }

    /* initialize dispatcher module */
    if (DispInit() == -1) {
        printf("Dispatch module initialization error\n");
        return -1;
    }
    
    /* parallel or serial */
    if (parallel == FALSE) {
        /* in this case single dissector that generate and insert one PEI call 'directly' the
           insert function of dispatcher module, otherwise a thread is the middelware from dissectors
           and dispatch function module */
        res = pthread_create(&pid, NULL, DispatchAgent, NULL);
        if (res != 0) {
            LogPrintf(LV_ERROR, "Dispatch parallel setup failed");
            return -1;
        }
        pthread_detach(pid);
    }

    return 0;
}


int PeiInit(pei *ppei)
{
    memset(ppei, 0, sizeof(pei));
    
    return 0;
}


int PeiIns(pei *ppei)
{
    /* end time */
    if (ppei != NULL)
        ppei->time = time(NULL);
    
    /* parallel/concurrent insert */
    if (parallel)
        return DispInsPei(ppei);
    
    /* push in the list */

    return 0;
}


int PeiFree(pei *ppei)
{
    pei_component *cmpn, *nxt;

    /* free components */
    cmpn = ppei->components;
    while (cmpn != NULL) {
        nxt = cmpn->next;
        if (cmpn->strbuf != NULL)
            DMemFree(cmpn->strbuf);
        if (cmpn->name != NULL)
            DMemFree(cmpn->name);
        if (cmpn->file_path != NULL)
            DMemFree(cmpn->file_path);
        DMemFree(cmpn);
        cmpn = nxt;
    }

    /* free protocol stack */
    if (ppei->stack != NULL)
        ProtDelFrame(ppei->stack);

    DMemFree(ppei);

    return 0;
}


void PeiPrint(pei *ppei)
{
    pei_component *cmpn;
    char tv[256];

    cmpn = ppei->components;
    while (cmpn != NULL) {
        LogPrintf(LV_INFO, "Pei component: %i", cmpn->id);
        LogPrintf(LV_INFO, "\teid: %i", cmpn->eid);
        if (cmpn->strbuf != NULL)
            LogPrintf(LV_INFO, "\tstr: %s", cmpn->strbuf);
        if (cmpn->file_path != NULL)
            LogPrintf(LV_INFO, "\tfile: %s", cmpn->file_path);
        if (cmpn->time_cap != 0) {
            strcpy(tv, ctime((time_t *)&cmpn->time_cap));
            tv[strlen(tv)-1] = '\0';
            LogPrintf(LV_INFO, "\ttime: %s", tv);
        }
        if (cmpn->time_cap_end != 0) {
            strcpy(tv, ctime((time_t *)&cmpn->time_cap_end));
            tv[strlen(tv)-1] = '\0';
            LogPrintf(LV_INFO, "\tend time: %s", tv);
        }
        if (cmpn->err != ELMT_ER_NONE)
            LogPrintf(LV_INFO, "\terr: %i", cmpn->err);
        cmpn = cmpn->next;
    }
}

